from llm_easy_tools import get_tool_defs
from typing import Annotated
from pprint import pprint

def contact_user(
        name: Annotated[str, "The name of the user"],
        email: Annotated[str, "The email of the user"],
        phone: Annotated[str, "The phone number of the user"]
        ) -> str:
    """
    Contact the user with the given name, email, and phone number.
    """
    pass


pprint(get_tool_defs([contact_user]))

#  OUPUT

[{
    'function': {
        'description': 'Contact the user with the given name, email, and phone number.',
        'name': 'contact_user',
        'parameters': {
            'properties': {
                'email': {
                    'description': 'The email of the user',
                    'type': 'string'
                },
                'name': {
                    'description': 'The name of the user',
                    'type': 'string'
                },
                'phone': {
                    'description': 'The phone number of the user',
                    'type': 'string'
                }
            },
            'required': ['name', 'email', 'phone'],
            'type': 'object'
        },
        'type': 'function'
    }
}]



import inspect
from typing import Annotated, Callable, Dict, Any, get_origin, Type, Union

import copy
import pydantic as pd
from pydantic import BaseModel
from pydantic_core import PydanticUndefined

from pprint import pprint


def llm_function(schema_name=None):
    def decorator(func):
        setattr(func, 'LLMEasyTools_external_function', True)
        if schema_name is not None:
            setattr(func, 'LLMEasyTools_schema_name', schema_name)
        return func
    return decorator



def tool_def(function_schema: dict) -> dict:
    return {
        "type": "function",
        "function": function_schema,
    }

def get_tool_defs(
        functions: list[Callable],
        case_insensitive: bool = False,
        prefix_class: Type[BaseModel] = None,
        prefix_schema_name: bool = True
        ) -> list[dict]:
    result = []
    for function in functions:
        fun_schema = get_function_schema(function, case_insensitive)
        if prefix_class:
            fun_schema = insert_prefix(prefix_class, fun_schema, prefix_schema_name, case_insensitive)
        result.append(tool_def(fun_schema))
    return result

def parameters_basemodel_from_function(function: Callable) -> Type[pd.BaseModel]:
    fields = {}
    parameters = inspect.signature(function).parameters
    for name, parameter in parameters.items():
        description = None
        type_ = parameter.annotation
        if get_origin(parameter.annotation) is Annotated:
            if parameter.annotation.__metadata__:
                description = parameter.annotation.__metadata__[0]
            type_ = parameter.annotation.__args__[0]
        default = PydanticUndefined if parameter.default is inspect.Parameter.empty else parameter.default
        fields[name] = (type_, pd.Field(default, description=description))
    return pd.create_model(f'{function.__name__}_ParameterModel', **fields)


def _recursive_purge_titles(d: Dict[str, Any]) -> None:
    """Remove a titles from a schema recursively"""
    if isinstance(d, dict):
        for key in list(d.keys()):
            if key == 'title' and "type" in d.keys():
                del d[key]
            else:
                _recursive_purge_titles(d[key])

def get_name(func: Callable, case_insensitive: bool = False) -> str:
    schema_name = func.__name__
    if hasattr(func, 'LLMEasyTools_schema_name'):
        schema_name = func.LLMEasyTools_schema_name
    if case_insensitive:
        schema_name = schema_name.lower()
    return schema_name


def get_function_schema(function: Callable, case_insensitive: bool=False) -> dict:
    function_schema = {
        'name': get_name(function, case_insensitive),
        'description': (function.__doc__ or '').strip(),
    }
    model = parameters_basemodel_from_function(function)
    if model.model_fields:
        model_json_schema = model.model_json_schema()
        _recursive_purge_titles(model_json_schema)
        function_schema['parameters'] = model_json_schema
    return function_schema

def insert_prefix(prefix_class, schema, prefix_schema_name=True, case_insensitive = False):
    if not issubclass(prefix_class, BaseModel):
        raise TypeError(
            f"The given class reference is not a subclass of pydantic BaseModel"
        )
    prefix_schema = prefix_class.model_json_schema()
    _recursive_purge_titles(prefix_schema)
    prefix_schema.pop('description', '')

    if 'parameters' in schema:
        prefix_schema['required'].extend(schema['parameters']['required'])
        for key, value in schema['parameters']['properties'].items():
            prefix_schema['properties'][key] = value
    new_schema = copy.copy(schema)  # Create a shallow copy of the schema
    new_schema['parameters'] = prefix_schema
    if len(new_schema['parameters']['properties']) == 0:  # If the parameters list is empty
        new_schema.pop('parameters')
    if prefix_schema_name:
        if case_insensitive:
            prefix_name = prefix_class.__name__.lower()
        else:
            prefix_name = prefix_class.__name__
        new_schema['name'] = prefix_name + "_and_" + schema['name']
    return new_schema


#######################################
#
# Examples

if __name__ == "__main__":
    def function_with_doc():
        """
        This function has a docstring and no parameteres.
        Expected Cost: high
        """
        pass

    @llm_function(schema_name="altered_name")
    def function_decorated():
        pass

    class ExampleClass:
        def simple_method(self, count: int, size: float):
            """simple method does something"""
            pass

    example_object = ExampleClass()

    class User(BaseModel):
        name: str
        age: int

    pprint(get_tool_defs([
        example_object.simple_method, 
        function_with_doc, 
        function_decorated,
        User
        ]))




from llm_easy_tools import get_tool_defs, process_response
from pydantic import BaseModel
from pprint import pprint

from pydantic import BaseModel
from pprint import pprint


class Address(BaseModel):
    street: str
    city: str

class Company(BaseModel):
    name: str
    speciality: str
    address: Address


def print_companies(companies: list[Company]):
    pprint(companies)
    return companies

#pprint(get_tool_defs([print_companies]))

file_path = 'examples/Three_Companies_Story.txt'
with open(file_path, 'r') as file:
    story = file.read()

content = f"{story}\n\nPlease print the information about companies mentioned in the text above."

#client = OpenAI()
from groq import Groq
client = Groq()

response = client.chat.completions.create(
    model="llama3-8b-8192",
    messages=[{"role": "user", "content": content}],
    tools=get_tool_defs([print_companies]),
    tool_choice="auto",
)


process_response(response, [print_companies])

# OUTPUT
[
    Company(
        name='Aether Innovations', 
        speciality='sustainable energy solutions', 
        address=Address(street='150 Futura Plaza', city='Metropolis')),
    Company(
        name='Gastronauts',
        speciality='culinary startup',
        address=Address(street='45 Flavor Street', city='Metropolis')),
    Company(
        name='SereneScape',
        speciality='digital wellness',
        address=Address(street='800 Tranquil Trail', city='Metropolis'))]




import json
from typing import Callable
from pprint import pprint
from typing import Type, Optional, List, Union
from pydantic import BaseModel, ValidationError

from openai.types.chat.chat_completion import ChatCompletionMessage, ChatCompletion, Choice
from openai.types.chat.chat_completion_message_tool_call   import ChatCompletionMessageToolCall, Function

from llm_easy_tools.schema_generator import get_function_schema, get_name, insert_prefix, tool_def, llm_function
from llm_easy_tools.processor import ToolResult, process_response as processor_process_response, get_toolset_tools

class ToolBox:
    def __init__(self,
                 fix_json_args: bool = True,
                 case_insensitive: bool = False,
                 insert_prefix_name: bool = True,
                 ):
        self._tool_registry: dict[str, dict[str, Callable | Type[BaseModel]]] = {}
        self._tool_sets: dict[str, object] = {}
        self.fix_json_args = fix_json_args
        self.case_insensitive = case_insensitive
        self.insert_prefix_name = insert_prefix_name


    def register_function(self, function: Callable) -> None:
        function_name = get_name(function, self.case_insensitive)

        if function_name in self._tool_registry:
            raise Exception(f"Trying to register {function_name} which is already registered")

        self._tool_registry[function_name] = { 'function': function }

    def register_toolset(self, obj: object, key=None) -> None:
        if key is None:
            key = type(obj).__name__

        if key in self._tool_sets:
            raise Exception(f"A toolset with key {key} already exists.")

        self._tool_sets[key] = obj
        for tool in get_toolset_tools(obj):
            self.register_function(tool)


    def get_toolset(self, key: str) -> object:
        return self._tool_sets[key]

    def tool_schemas(self, prefix_class=None, predicate=None) -> list[dict]:
        if predicate is None:
            predicate = lambda _: True
        schemas = []
        for tool_name, tool_value in self._tool_registry.items():
            if not predicate(tool_value):
                continue
            schemas.append(self.get_tool_schema(tool_name, prefix_class))
        return schemas


    def get_tool_schema(self, tool_name: str, prefix_class=None) -> dict:
        tool = self._tool_registry[tool_name]
        the_schema = get_function_schema(tool['function'])
        if prefix_class is not None:
            the_schema = insert_prefix(prefix_class, the_schema, self.insert_prefix_name, self.case_insensitive)
        return tool_def(the_schema)

    def process_response(self, response: ChatCompletion, choice_num=0, prefix_class=None) -> list[ToolResult]:
        functions = [f['function'] for f in self._tool_registry.values()]
        return  processor_process_response(response, functions, choice_num, prefix_class)
    


#######################################
#
# Examples


if __name__ == "__main__":
    @llm_function(schema_name="altered_name")
    def function_decorated():
        return 'Result of function_decorated'

    class ExampleClass:
        def simple_method(self, count: int, size: float):
            """simple method does something"""
            return 'Result of simple_method'

    example_object = ExampleClass()

    toolbox = ToolBox()
    toolbox.register_function(function_decorated)
    toolbox.register_function(example_object.simple_method)
#    pprint(generate_tools(function_with_doc, function_decorated))
    pprint(toolbox.tool_schemas())

    chat_completion_message = ChatCompletionMessage(
        role="assistant",
        tool_calls=[
            ChatCompletionMessageToolCall( 
                id ='A',
                type = 'function',
                function = Function(
                    arguments = json.dumps({"count": 1, "size": 2.2}),
                    name = 'simple_method'
                )
            )
        ]
    )

    chat_completion = ChatCompletion(
        id='A',
        created=0,
        model='A',
        choices=[Choice(finish_reason='stop', index=0, message=chat_completion_message)],
        object='chat.completion'
    )

    print(toolbox.process_response(chat_completion))






from .tool_box import ToolBox, ToolResult
from .schema_generator import get_function_schema, llm_function, insert_prefix, get_tool_defs
from .processor import process_response, process_tool_call, get_toolset_tools



from llm_easy_tools import get_tool_defs, process_response
from openai import OpenAI
from pprint import pprint

client = OpenAI()

def contact_user(name: str, city: str) -> str:
    return f"User {name} from {city} was contactd"

response = client.chat.completions.create(
    model="gpt-3.5-turbo-1106",
    messages=[{"role": "user", "content": "Contact John. John lives in Warsaw"}],
    tools=get_tool_defs([contact_user]),
    tool_choice={"type": "function", "function": {"name": "contact_user"}},
)
# There might be more than one tool calls in a single response so results are a list
results = process_response(response, [contact_user])

pprint(results[0].output)



import pytest
import json

from unittest.mock import Mock
from llm_easy_tools import ToolBox, ToolResult, llm_function
from pydantic import BaseModel, Field, ValidationError
from typing import Any
from openai.types.chat.chat_completion import ChatCompletionMessage, ChatCompletion

class TestTool:

    class SomeClass(BaseModel):
        value: int

    @llm_function()
    def tool_method(self, arg: int) -> str:
        return f'executed tool_method with param: {arg}'

    @llm_function()
    def additional_tool_method(self, arg: int) -> str:
        return f'executed additional_tool_method with param: {arg}'

    def _private_tool_method(self, arg: int) -> str:
        return str(arg.value * 4)

    @llm_function()
    class User(BaseModel):
        name: str
        age: int

    @llm_function('short_address')
    class Address(BaseModel):
        city: str
        street: str

    @llm_function()
    def no_output(self, arg: int):
        pass


    @llm_function()
    def failing_method(self, arg: int) -> str:
        raise Exception('Some exception')



tool = TestTool()

def test_toolbox_init():
    toolbox = ToolBox()
    assert toolbox._tool_registry == {}
    assert toolbox.tool_schemas() == []
    assert toolbox.case_insensitive == False
    assert toolbox.fix_json_args == True

def test_register_toolset():
    tool_manager = ToolBox()

    # Test the normal case
    tool_manager.register_toolset(tool)

    toolset = tool_manager.get_toolset('TestTool')
    assert toolset == tool

    assert 'tool_method' in tool_manager._tool_registry
    assert 'additional_tool_method' in tool_manager._tool_registry
    assert 'User' in tool_manager._tool_registry
    assert 'SomeClass' not in tool_manager._tool_registry
    assert 'short_address' in tool_manager._tool_registry
    assert '_private_tool_method' not in tool_manager._tool_registry

    # Test for Exception when a Toolset with same key is being registered
    with pytest.raises(Exception) as exception_info:
        tool_manager.register_toolset(tool)

    assert str(exception_info.value) == 'A toolset with key TestTool already exists.'


class FunctionCallMock:
    def __init__(self, name, arguments):
        self.name = name
        self.arguments = arguments



class UserDetail(BaseModel):
    name: str
    age: int

def mk_chat_with_tool_call(name, args):
    message = ChatCompletionMessage(
        role="assistant",
        tool_calls=[
            {
                "id": 'A',
                "type": 'function',
                "function": {
                    "arguments": json.dumps(args),
                    "name": name
                }
            }
        ]
    )
    chat_completion = ChatCompletion(
        id='A',
        created=0,
        model='A',
        choices=[{'finish_reason': 'stop', 'index': 0, 'message': message}],
        object='chat.completion'
    )
    return chat_completion


def test_process_response():
    # too much mocking in this test
    toolbox = ToolBox()
    toolbox.register_function(UserDetail)
    original_user = UserDetail(name="John", age=21)
    response = mk_chat_with_tool_call("UserDetail", original_user.model_dump())
    results = toolbox.process_response(response)
    assert len(results) == 1
    assert results[0].output == original_user
    message = results[0].to_message()
    assert message['role'] == 'tool'
    assert message['tool_call_id'] == 'A'
    assert message['name'] == 'UserDetail'
    assert message['content'] == 'UserDetail created'


# Define the test cases
def test_register_tool():

    def example_tool(name: str):
        print(f'Running test tool with name param: "{name}"')


    toolbox = ToolBox()

    # Test with correct parameters
    toolbox.register_function(example_tool)
    assert 'example_tool' in toolbox._tool_registry
    function_info = toolbox._tool_registry['example_tool']
    assert function_info["function"] == example_tool
    assert len(toolbox.tool_schemas()) == 1

    # Test with function with no parameters
    def no_parameters(): pass
    toolbox.register_function(no_parameters)
    assert 'no_parameters' in toolbox._tool_registry

    @llm_function("good_name")
    def bad_name_tool(name: str):
        print('Running bad_name_tool')

    toolbox.register_function(bad_name_tool)
    assert 'good_name' in toolbox._tool_registry
    function_info = toolbox._tool_registry['good_name']
    assert function_info["function"] == bad_name_tool

def test_register_model():

    class Tool(BaseModel):
        name: str

    class WikiSearch(BaseModel):
        query: str

    toolbox = ToolBox()
    toolbox.register_function(Tool)
    x = toolbox._tool_registry['Tool']['function'](name="Some name")
    assert x.name == 'Some name'

    assert len(toolbox.tool_schemas()) == 1
    assert toolbox.tool_schemas()[0]['function']['name'] == 'Tool'

    toolbox.register_function(WikiSearch)
    assert len(toolbox.tool_schemas()) == 2

    assert toolbox.get_tool_schema('WikiSearch')['function']['name'] == 'WikiSearch'

def test_prefixing():

    class Reflection(BaseModel):
        relevancy: str = Field(..., description="Whas the last retrieved information relevant and why?")

    def example_tool(name: str):
        return 'test tool result'

    toolbox = ToolBox()
    toolbox.register_function(example_tool)
    tool_schemas = toolbox.tool_schemas(prefix_class=Reflection)
    assert len(tool_schemas) == 1
    function_schema = tool_schemas[0]['function']
    first_param_name = list(function_schema['parameters']['properties'].keys())[0]
    assert first_param_name == 'relevancy'
    assert function_schema['name'] == 'Reflection_and_example_tool'


def test_process_response_with_prefixing():

    class Reflection(BaseModel):
        relevancy: str = Field(..., description="Whas the last retrieved information relevant and why?")

    toolbox = ToolBox()
    toolbox.register_toolset(tool)
    prefixed_name = 'Reflection_and_tool_method'
    response = mk_chat_with_tool_call(prefixed_name, {'arg': 2})
    results = toolbox.process_response(response, prefix_class=Reflection)
    assert results[0].output == 'executed tool_method with param: 2'
    assert results[0].prefix is None
    assert len(results[0].soft_errors) == 1
    args = {'arg': 2}
    args['relevancy'] = 'very good'
    response = mk_chat_with_tool_call(prefixed_name, args)
    results = toolbox.process_response(response, prefix_class=Reflection)
    assert results[0].output == 'executed tool_method with param: 2'
    assert isinstance(results[0].prefix, Reflection)

pytest.main()



import pytest

from typing import List, Optional, Union, Literal, Annotated
from pydantic import BaseModel, Field, field_validator

from llm_easy_tools import get_function_schema, insert_prefix, llm_function 

from llm_easy_tools.schema_generator import parameters_basemodel_from_function, _recursive_purge_titles, get_name

from pprint import pprint


def simple_function(count: int, size: Optional[float] = None):
    """simple function does something"""
    pass


def simple_function_no_docstring(
        apple: Annotated[str, 'The apple'],
        banana: Annotated[str, 'The banana']
):
    pass




def test_function_schema():

    function_schema = get_function_schema(simple_function)
    assert function_schema['name'] == 'simple_function'
    assert function_schema['description'] == 'simple function does something'
    params_schema = function_schema['parameters']
    assert len(params_schema['properties']) == 2
    assert params_schema['type'] == "object"
    assert params_schema['properties']['count']['type'] == "integer"
    assert 'size' in params_schema['properties']
    assert 'title' not in params_schema
    assert 'title' not in params_schema['properties']['count']
    assert 'description' not in params_schema

def test_noparams():
    def function_with_no_params():
        """
        This function has a docstring and takes no parameters.
        """
        pass

    def function_no_doc():
        pass

    result = get_function_schema(function_with_no_params)
    assert result['name'] == 'function_with_no_params'
    assert result['description'] == "This function has a docstring and takes no parameters."
    assert 'parameters' not in result # "Omitting parameters defines a function with an empty parameter list."

    # Function without docstring and EmptyModel as parameter
    result = get_function_schema(function_no_doc)
    assert result['name'] == 'function_no_doc'
    assert result['description'] == ''
    assert 'parameters' not in result


def test_nested():
    class Foo(BaseModel):
        count: int
        size: Optional[float] = None

    class Bar(BaseModel):
        """Some Bar"""
        apple: str = Field(description="The apple")
        banana: str = Field(description="The banana")

    class FooAndBar(BaseModel):
        foo: Foo
        bar: Bar

    def nested_structure_function(foo: Foo, bars: List[Bar]):
        """spams everything"""
        pass

    function_schema = get_function_schema(nested_structure_function)
    assert function_schema['name'] == 'nested_structure_function'
    assert function_schema['description'] == 'spams everything'
    assert len(function_schema['parameters']['properties']) == 2

    function_schema = get_function_schema(FooAndBar)
    assert function_schema['name'] == 'FooAndBar'
    assert len(function_schema['parameters']['properties']) == 2


def test_methods():
    class ExampleClass:
        def simple_method(self, count: int, size: Optional[float] = None):
            """simple method does something"""
            pass

    example_object = ExampleClass()

    function_schema = get_function_schema(example_object.simple_method)
    assert function_schema['name'] == 'simple_method'
    assert function_schema['description'] == 'simple method does something'
    params_schema = function_schema['parameters']
    assert len(params_schema['properties']) == 2

def test_name_change():
    @llm_function('changed_name')
    def new_simple_function(count: int, size: Optional[float] = None):
        """simple function does something"""
        pass

    function_schema = get_function_schema(new_simple_function)
    assert function_schema['name'] == 'changed_name'


def test_merge_schemas():

    class Reflection(BaseModel):
        relevancy: str = Field(..., description="Whas the last retrieved information relevant and why?")
        next_actions_plan: str = Field(..., description="What you plan to do next and why")

    function_schema = get_function_schema(simple_function)
    new_schema = insert_prefix(Reflection, function_schema)
    assert new_schema['name'] == "Reflection_and_simple_function"
    assert len(new_schema['parameters']['properties']) == 4
    assert len(new_schema['parameters']['required']) == 3
    assert len(function_schema['parameters']['properties']) == 2  # the old schema is not changed
    assert len(function_schema['parameters']['required']) == 1  # the old schema is not changed
    param_names = list(new_schema['parameters']['properties'].keys())
    assert param_names == ['relevancy', 'next_actions_plan', 'count', 'size']

    function_schema = get_function_schema(simple_function)
    new_schema = insert_prefix(Reflection, function_schema, case_insensitive=True)
    assert new_schema['name'] == "reflection_and_simple_function"

def test_noparams_function_merge():

    def function_no_params():
        pass

    class Reflection(BaseModel):
        relevancy: str = Field(..., description="Whas the last retrieved information relevant and why?")
        next_actions_plan: str = Field(..., description="What you plan to do next and why")

    # Function without docstring and EmptyModelWithDoc as parameter
    function_schema = get_function_schema(function_no_params)
    assert function_schema['name'] == 'function_no_params'
    assert 'parameters' not in function_schema

    new_schema = insert_prefix(Reflection, function_schema)
    assert len(new_schema['parameters']['properties']) == 2
    assert new_schema['name'] == 'Reflection_and_function_no_params'

def test_model_init_function():

    class User(BaseModel):
        """A user object"""
        name: str
        city: str


    function_schema = get_function_schema(User)
    assert function_schema['name'] == 'User'
    assert function_schema['description'] == 'A user object'
    assert len(function_schema['parameters']['properties']) == 2
    assert len(function_schema['parameters']['required']) == 2


def test_case_insensitivity():

    class User(BaseModel):
        """A user object"""
        name: str
        city: str

    function_schema = get_function_schema(User, case_insensitive=True)
    assert function_schema['name'] == 'user'
    assert get_name(User, case_insensitive=True) == 'user'




from llm_easy_tools import get_tool_defs, process_response
from pydantic import BaseModel
from openai import OpenAI
from pprint import pprint

client = OpenAI()

# Define a Pydantic model for your tool's output
class UserDetail(BaseModel):
    name: str
    city: str


response = client.chat.completions.create(
    model="gpt-3.5-turbo-1106",
    messages=[{"role": "user", "content": "Extract user details from the following sentence: John lives in Warsaw and likes banana"}],
    tools=get_tool_defs([UserDetail]),
    tool_choice="auto",
)
# There might be more than one tool calls in a single response so results are a list
results = process_response(response, [UserDetail])

#pprint(results)
pprint(results[0].output)




import json

from openai.types.chat.chat_completion import ChatCompletion, ChatCompletionMessage, Choice
from pprint import pprint

chat_completion_message = ChatCompletionMessage(
    role="assistant",
    tool_calls=[
        {
            "id": 'A',
            "type": 'function',
            "function": {
                "arguments": json.dumps({"param1": "value1", "param2": "value2"}),
                "name": 'function_name'
            }
        }
    ]
)


pprint(chat_completion_message)

choice = Choice(
    finish_reason="stop",
    index=0,
    message=chat_completion_message
)

response = ChatCompletion(
    id='A',
    created=0,
    model='A',
    choices=[ choice ],
    object='chat.completion'
)

pprint(response)


import pytest
import json
from time import sleep, time

from unittest.mock import Mock
from llm_easy_tools import llm_function
from pydantic import BaseModel, Field, ValidationError
from typing import Any
from openai.types.chat.chat_completion import ChatCompletionMessage, ChatCompletion, Choice
from openai.types.chat.chat_completion_message_tool_call   import ChatCompletionMessageToolCall, Function

from llm_easy_tools.processor import process_response, process_tool_call, ToolResult, _extract_prefix_unpacked

def mk_tool_call(name, args):
    arguments = json.dumps(args)
    return ChatCompletionMessageToolCall(id='A', function=Function(name=name, arguments=arguments), type='function')

def mk_tool_call_jason(name, args):
    return ChatCompletionMessageToolCall(id='A', function=Function(name=name, arguments=args), type='function')

def mk_chat_completion(tool_calls):
    return ChatCompletion(
        id='A',
        created=0,
        model='gpt-3.5-turbo',
        object='chat.completion',
        choices=[
            Choice(
                finish_reason='stop',
                index=0,
                message=ChatCompletionMessage(role='assistant', tool_calls=tool_calls))
        ]
    )


def test_process_methods():
    class TestTool:

        def tool_method(self, arg: int) -> str:
            return f'executed tool_method with param: {arg}'

        def no_output(self, arg: int):
            pass

        def failing_method(self, arg: int) -> str:
            raise Exception('Some exception')


    tool = TestTool()

    tool_call = mk_tool_call("tool_method", {"arg": 2})
    result = process_tool_call(tool_call, [tool.tool_method])
    assert isinstance(result, ToolResult)
    assert result.output == 'executed tool_method with param: 2'

    tool_call = mk_tool_call("failing_method", {"arg": 2})
    result = process_tool_call(tool_call, [tool.failing_method])
    assert isinstance(result, ToolResult)
    assert "Some exception" in str(result.error)
    message = result.to_message()
    assert "Some exception" in message['content']

    tool_call = mk_tool_call("no_output", {"arg": 2})
    result = process_tool_call(tool_call, [tool.no_output])
    assert isinstance(result, ToolResult)
    message = result.to_message()
    assert message['content'] == ''

def test_process_complex():

    class Address(BaseModel):
        street: str
        city: str

    class Company(BaseModel):
        name: str
        speciality: str
        address: Address


    def print_companies(companies: list[Company]):
        return companies

    company_list = [{
        'address': {'city': 'Metropolis', 'street': '150 Futura Plaza'},
        'name': 'Aether Innovations',
        'speciality': 'sustainable energy solutions'
    }]

    tool_call = mk_tool_call("print_companies", {"companies": company_list})
    result = process_tool_call(tool_call, [print_companies])
    assert isinstance(result, ToolResult)
    assert isinstance(result.output, list)
    assert isinstance(result.output[0], Company)

def test_prefixing():

    class Reflection(BaseModel):
        relevancy: str = Field(..., description="Whas the last retrieved information relevant and why?")

    args = { 'relevancy': 'good', 'name': 'hammer'}
    prefix = _extract_prefix_unpacked(args, Reflection)
    assert isinstance(prefix, Reflection)
    assert 'reflection' not in args # prefix params extracted


def test_json_fix():

    class UserDetail(BaseModel):
        name: str
        age: int

    original_user = UserDetail(name="John", age=21)
    json_data = json.dumps(original_user.model_dump())
    json_data = json_data[:-1]
    json_data = json_data + ',}'
    tool_call = mk_tool_call_jason("UserDetail", json_data)
    result = process_tool_call(tool_call, [UserDetail])
    assert result.output == original_user
    assert len(result.soft_errors) > 0

    result = process_tool_call(tool_call, [UserDetail], fix_json_args=False)
    assert isinstance(result.error, json.decoder.JSONDecodeError)

    response = mk_chat_completion([tool_call])
    results = process_response(response, [UserDetail])
    assert results[0].output == original_user
    assert len(results[0].soft_errors) > 0

    results = process_response(response, [UserDetail], fix_json_args=False)
    assert isinstance(results[0].error, json.decoder.JSONDecodeError)

def test_case_insensitivity():
    class User(BaseModel):
        name: str
        city: str

    response = mk_chat_completion([mk_tool_call("user", {"name": "John", "city": "Metropolis"})])
    results = process_response(response, [User], case_insensitive=True)
    assert results[0].output == User(name="John", city="Metropolis")

def test_parallel_tools():

    class CounterClass:
        def __init__(self):
            self.counter = 0

        def increment_counter(self):
            self.counter += 1
            sleep(1)  # Increased sleep time to 1 second

    counter = CounterClass()
    tool_call = mk_tool_call("increment_counter", {})
    response = mk_chat_completion([tool_call] * 10)

    start_time = time()
    results = process_response(response, [counter.increment_counter])
    end_time = time()

    assert results[9].error is None

    time_taken = end_time - start_time
    assert counter.counter == 10
    assert time_taken <= 2, f"Expected processing time to be less than or equal to 2 seconds, but was {time_taken}"



from llm_easy_tools import llm_function, get_toolset_tools, get_tool_defs, process_response
import re
from groq import Groq
client = Groq()

# Define a class for document management
class DocumentManager:
    def __init__(self):
        self.documents = [
            {"title": "Introduction to GPT-4", "content": "GPT-4 is the latest language model by OpenAI. It has several new features."},
            {"title": "Advancements in AI", "content": "Exploring the latest advancements in artificial intelligence. AI is evolving rapidly."},
            {"title": "Natural Language Processing", "content": "Understanding NLP and its applications. NLP is a key part of AI."}
        ]
        self.current_document = None

    @llm_function()
    def search_document(self, term: str):
        closest_match = None
        min_distance = float('inf')
        for doc in self.documents:
            distance = self._calculate_similarity(term, doc['title'])
            if distance < min_distance:
                min_distance = distance
                closest_match = doc

        if closest_match:
            self.current_document = closest_match
            first_sentence = closest_match['content'].split('.')[0]
            return f"Found document, title: {closest_match['title']}.\n {first_sentence}..."
        else:
            return "No matching document found."

    @llm_function()
    def lookup_word(self, word: str):
        if not self.current_document:
            return "No document is currently selected."

        sentences = self.current_document['content'].split('.')
        for sentence in sentences:
            if word in sentence:
                return sentence.strip()
        return f"The word '{word}' was not found in the current document."

    def _calculate_similarity(self, term, title):
        # Simple similarity calculation (can be replaced with more sophisticated methods)
        return abs(len(term) - len(title))  # Example simplistic similarity measure

# Create an instance of DocumentManager
doc_manager = DocumentManager()
tools = get_toolset_tools(doc_manager)

# Example LLM call to search for a document
response_search = client.chat.completions.create(
    model="llama3-8b-8192",
    messages=[{"role": "user", "content": "Find a document about AI"}],
    tools=get_tool_defs(tools),
    tool_choice="auto"
)

# Process the response to search for the document

results_search = process_response(response_search, tools)
print(results_search[0].output)

# Example LLM call to look up a word in the current document
response_lookup = client.chat.completions.create(
    model="llama3-8b-8192",
    messages=[{"role": "user", "content": "Look up the word 'evolving'"}],
    tools=get_tool_defs(tools),
    tool_choice="auto",
)
# Process the response to look up the word
results_lookup = process_response(response_lookup, get_toolset_tools(doc_manager))
print(results_lookup[0].output)


## OUTPUT
#
# Found document, title: Advancements in AI.
# Exploring the latest advancements in artificial intelligence...
# AI is evolving rapidly



import json
import traceback
import inspect

from concurrent.futures import ThreadPoolExecutor
from typing import Callable
from pprint import pprint
from typing import Type, Optional, List, Union
from pydantic import BaseModel, ValidationError
from dataclasses import dataclass, field

from openai.types.chat.chat_completion import ChatCompletionMessage, ChatCompletion, Choice
from openai.types.chat.chat_completion_message_tool_call   import ChatCompletionMessageToolCall, Function

from llm_easy_tools.schema_generator import get_name, llm_function, parameters_basemodel_from_function

class NoMatchingTool(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


@dataclass
class ToolResult:
    """
    Represents the result of a tool invocation within the ToolBox framework.

    Attributes:
        tool_call_id (str): A unique identifier for the tool call.
        name (str): The name of the tool that was called.
        output (Optional[Union[str, BaseModel]]): The output generated by the tool call, if any.
        error (Optional[Exception]): An error message if the tool call failed.
        soft_errors (List[Exception]): A list of non-critical error messages encountered during the tool call.
        prefix (Optional[BaseModel]): The Pydantic model instance used as a prefix in the tool call, if applicable.

    Methods:
        to_message(): Converts the ToolResult into a dictionary suitable for returning to a chat interface.
    """
    tool_call_id: str
    name: str
    output: Optional[Union[str, BaseModel]] = None
    error: Optional[Exception] = None
    soft_errors: List[Exception] = field(default_factory=list)
    prefix: Optional[BaseModel] = None

    def to_message(self) -> dict[str, str]:
        if self.error is not None:
            content = f"{self.error}"
        elif self.output is None:
            content = ''
        elif isinstance(self.output, str):
            content = self.output
        elif isinstance(self.output, BaseModel):
            content = f"{self.name} created"
        else:
            content = self.output.to_content()
        return {
            "role": "tool",
            "tool_call_id": self.tool_call_id,
            "name": self.name,
            "content": content,
        }

def process_tool_call(tool_call, functions_or_models, prefix_class=None, fix_json_args=True, case_insensitive=False) -> ToolResult:
    function_call = tool_call.function
    tool_name = function_call.name
    args = function_call.arguments
    soft_errors = []
    error = None
    prefix = None
    output = None
    try:
        tool_args = json.loads(args)
    except json.decoder.JSONDecodeError as e:
        if fix_json_args:
            soft_errors.append(e)
            args = args.replace(', }', '}').replace(',}', '}')
            tool_args = json.loads(args)
        else:
            return ToolResult(tool_call_id=tool_call.id, name=tool_name, error=e)

    if prefix_class is not None:
        try:
            prefix = _extract_prefix_unpacked(tool_args, prefix_class)
        except ValidationError as e:
            soft_errors.append(e)
        prefix_name = prefix_class.__name__
        if case_insensitive:
            prefix_name = prefix_name.lower()
        if not tool_name.startswith(prefix_name):
            soft_errors.append(NoMatchingTool(f"Trying to decode function call with a name '{tool_name}' not matching prefix '{prefix_name}'"))
        else:
            tool_name = tool_name[len(prefix_name + '_and_'):]
    function_found = False
    for f in functions_or_models:
        if get_name(f, case_insensitive=case_insensitive) == tool_name:
            function_found = True
            try:
                output = _process_unpacked(f, tool_args)
            except Exception as e:
                error = e
            break
    if not function_found:
        error = NoMatchingTool(f"Function {tool_name} not found")
    result = ToolResult(
        tool_call_id=tool_call.id, 
        name=tool_name, 
        output=output, 
        error=error,
        soft_errors=soft_errors,
        prefix=prefix
    )
    return result

def _process_unpacked(function, tool_args={}) -> Union[str, BaseModel]:
    model = parameters_basemodel_from_function(function)
    model_instance = model(**tool_args)
    args = {}
    for field, _ in model.model_fields.items():
        args[field] = getattr(model_instance, field)
    return function(**args)

def _extract_prefix_unpacked(tool_args, prefix_class):
    # modifies tool_args
    prefix_args = {}
    for key in list(tool_args.keys()):  # copy keys to list because we modify the dict while iterating over it
        if key in prefix_class.__annotations__:
            prefix_args[key] = tool_args.pop(key)
    prefix = prefix_class(**prefix_args)
    return(prefix)

def process_response(
        response: ChatCompletion,
        functions: List[Callable],
        choice_num=0,
        prefix_class=None,
        fix_json_args=True,
        case_insensitive=False
        ) -> list[ToolResult]:
    """
    Processes a ChatCompletion response, executing contained tool calls.
    For each tool call matches a function from the 'functions' list by name.
    The result of the tool call is returned as a ToolResult object.
    If the tool call raises an exception, that exception is saved in the 'error' field in the result.

    Args:
        response (ChatCompletion): The response object containing tool calls.
        functions (List[Callable]): A list of functions or pydantic models to call.
        choice_num (int, optional): The index of the choice to process from the response. Defaults to 0.

    Returns:
        list[ToolResult]: A list of ToolResult objects, each representing the outcome of a processed tool call.
    """
    results = []
    if hasattr(response.choices[choice_num].message, 'function_call') and response.choices[choice_num].message.function_call:
        # this is obsolete in openai - but maybe it is used by other llms?
        function_call = response.choices[choice_num].message.function_call
        tool_calls = [ChatCompletionMessageToolCall(id='A', function=Function(name=function_call.name, arguments=function_call.arguments), type='function')]
    elif hasattr(response.choices[choice_num].message, 'tool_calls') and response.choices[choice_num].message.tool_calls:
        tool_calls = response.choices[choice_num].message.tool_calls
    else:
        tool_calls = []
        # Prepare the arguments for each tool call
    args_list = [(tool_call, functions, prefix_class, fix_json_args, case_insensitive) for tool_call in tool_calls]

    with ThreadPoolExecutor() as executor:
        results = list(executor.map(lambda args: process_tool_call(*args), args_list))
    return results

def get_toolset_tools(obj: object) -> list[Callable]:
    result = []
    methods = inspect.getmembers(obj, predicate=inspect.ismethod)
    for name, method in methods:
        if hasattr(method, 'LLMEasyTools_external_function'):
            result.append(method)
    for attr_name in dir(obj.__class__):
        attr_value = getattr(obj.__class__, attr_name)
        if isinstance(attr_value, type) and hasattr(attr_value, 'LLMEasyTools_external_function'):
            result.append(attr_value)
    return result



#######################################
#
# Examples


if __name__ == "__main__":

    @llm_function(schema_name="altered_name")
    def function_decorated():
        return 'Result of function_decorated'

    class ExampleClass:
        def simple_method(self, count: int, size: float):
            """simple method does something"""
            return 'Result of simple_method'

    example_object = ExampleClass()

    class User(BaseModel):
        name: str
        email: str

    def mk_chat_with_tool_call(name, args):
        message = ChatCompletionMessage(
            role="assistant",
            tool_calls=[
                {
                    "id": 'A',
                    "type": 'function',
                    "function": {
                        "arguments": json.dumps(args),
                        "name": name
                    }
                }
            ]
        )
        chat_completion = ChatCompletion(
            id='A',
            created=0,
            model='A',
            choices=[{'finish_reason': 'stop', 'index': 0, 'message': message}],
            object='chat.completion'
        )
        return chat_completion


    call_to_altered_name = mk_chat_with_tool_call('altered_name', {}).choices[0].message.tool_calls[0]
    pprint(call_to_altered_name)
    pprint(process_tool_call(call_to_altered_name, [function_decorated]))

    call_to_simple_method = mk_chat_with_tool_call('simple_method', {"count": 1, "size": 2.2}).choices[0].message.tool_calls[0]
    pprint(process_tool_call(call_to_simple_method, [example_object.simple_method]))

    call_to_model = mk_chat_with_tool_call('User', {"name": 'John', "email": 'john@example.com'}).choices[0].message.tool_calls[0]
    pprint(process_tool_call(call_to_model, [User]))




from pprint import pprint

error = {'error': {'message': "Failed to call a function. Please adjust your prompt. See 'failed_generation' for more details.", 'type': 'invalid_request_error', 'code': 'tool_use_failed', 'failed_generation': '```\n{\n  "tool_calls": [\n    {\n      "id": "pending",\n      "type": "function",\n      "function": {\n        "name": "print_companies"\n      },\n      "parameters": {\n        "companies": [\n          {\n            "name": "Aether Innovations",\n            "address": {\n              "street": "150 Futura Plaza"\n            },\n            "speciality": "sustainable energy solutions"\n          },\n          {\n            "name": "Gastronauts",\n            "address": {\n              "street": "45 Flavor Street"\n            },\n            "speciality": "urban dining experiences"\n          },\n          {\n            "name": "SereneScape",\n            "address": {\n              "street": "800 Tranquil Trail"\n            },\n            "speciality": "digital wellness"\n          }\n        ]\n      }\n    }\n  ]\n}\n```'}}

pprint(error)


from llm_easy_tools import get_tool_defs, process_response 
from llm_easy_tools.schema_generator import get_name, llm_function

from pydantic import BaseModel
from openai import OpenAI
from pprint import pprint

from groq import Groq
client = Groq()

# with Llama the function name needs to be better fit the action
@llm_function("extract_user_details")
class UserDetail(BaseModel):
    name: str
    city: str

response = client.chat.completions.create(
    model="llama3-8b-8192",
    messages=[{"role": "user", "content": "Extract user details from the following sentence: John lives in Warsaw and likes banana"}],
    tools=get_tool_defs([UserDetail]),
    tool_choice="auto"
)
# There might be more than one tool calls in a single response so results are a list
results = process_response(response, [UserDetail])

# #pprint(results)
pprint(results[0].output)
